/*
 * - config.c
 *
 * Read and prase the configuration files
 */

#include "seal.h"

static int CONFIG_get_section(char const* section, FILE* fp);
static int CONFIG_get_next_key_value(char* key, char* value, FILE* fp);
static void CONFIG_trim(char const* orig_str, char* ret_str);
static void CONFIG_lowercase(char* str);

/*
 * Prase the general section
 * Set the variables (uid, spread_ip, spread_port)
 */
void CONFIG_parse_general(char* uid, char* spread_ip, int* spread_port) {
	
	FILE* fp;				// conf file
	int is_uid_found = 0;			// flag tracing if the variables are set
	int is_spread_ip_found = 0; 
	int is_spread_port_found = 0;

	char key[CONFIG_MAX_CHAR];
	char value[CONFIG_MAX_CHAR];

	// open the config file
	if ((fp = fopen(CONFIG_FILE, "r")) == NULL) {
		debug_err("Unable to open config file %s\n", CONFIG_FILE);
		exit(-1);
	}

	// read the general section, and put the fp to the correct file location
	if (CONFIG_get_section("general", fp) == 0) {
		debug_err("Section [general] is not found\n");
		exit(-1);
	}

	// set global variables for matched keys
	memset(key, 0, CONFIG_MAX_CHAR);
	memset(value, 0, CONFIG_MAX_CHAR);
	while (CONFIG_get_next_key_value(key, value, fp)) {
		if (!strncmp(key, "uid", CONFIG_MAX_CHAR)) {
			memset(uid, 0, SEAL_UID_LEN);
			strncpy(uid, value, SEAL_UID_LEN);
			is_uid_found = 1;
		} else if (!strncmp(key, "spread_ip", CONFIG_MAX_CHAR)) {
			memset(spread_ip, 0, SEAL_IP_LEN);
			strncpy(spread_ip, value, SEAL_IP_LEN);
			is_spread_ip_found = 1;
		} else if (!strncmp(key, "spread_port", CONFIG_MAX_CHAR)) {
			*spread_port = atoi(value);
			is_spread_port_found = 1;
		} 
		memset(key, 0, CONFIG_MAX_CHAR);
		memset(value, 0, CONFIG_MAX_CHAR);
	}
	fclose(fp);

	if (!(is_uid_found && is_spread_ip_found && is_spread_port_found)) {
		debug_err("uid/spread_ip/spread_port is not found\n");
		exit(-1);
	}
}

/*
 * Parse the group section.
 * Retrieve group member uids' and store them in mem_uids. 
 * Return the number of group members (should be at least 1), -1 on error.
 */
int CONFIG_parse_group(char const* group_name, char** mem_uids) {

	FILE* fp;				// conf file
	char key[CONFIG_MAX_CHAR];
	char value[CONFIG_MAX_CHAR];
	int n = 0;

	// open the config file
	if ((fp = fopen(CONFIG_FILE, "r")) == NULL) {
		debug_err("Unable to open config file %s\n", CONFIG_FILE);
		return -1;
	}

	// read the general section, and put the fp to the correct file location
	if (CONFIG_get_section(group_name, fp) == 0) {
		debug_err("Section [%s] is not found\n", group_name);
		return -1;
	}

	// set the group member names
	memset(key, 0, CONFIG_MAX_CHAR);
	memset(value, 0, CONFIG_MAX_CHAR);
	while (CONFIG_get_next_key_value(key, value, fp)) {
		memset(mem_uids[n], 0, SEAL_UID_LEN);
		strncpy(mem_uids[n++], value, SEAL_UID_LEN);
			
		memset(key, 0, CONFIG_MAX_CHAR);
		memset(value, 0, CONFIG_MAX_CHAR);
	}

	// close the file
	fclose(fp);

	return n;
}

/*
 * Parse the size of the selected group
 * Return -1 on error.
 */
int CONFIG_parse_group_size(char const* group_name) {

	FILE* fp;				// conf file
	int n = 0;
	char dummy_key[CONFIG_MAX_CHAR];
	char dummy_value[CONFIG_MAX_CHAR];

	// open the config file
	if ((fp = fopen(CONFIG_FILE, "r")) == NULL) {
		debug_err("Unable to open config file %s\n", CONFIG_FILE);
		return -1;
	}

	// read the general section, and put the fp to the correct file location
	if (CONFIG_get_section(group_name, fp) == 0) {
		debug_err("Section [%s] is not found\n", group_name);
		return -1;
	}

	// set the group member names
	while (CONFIG_get_next_key_value(dummy_key, dummy_value, fp)) {
		n++;	
	}

	// close the file
	fclose(fp);

	return n;
}

/*---------------------------------------------------------------------
 *                       STATIC FUNCTIONS
 *---------------------------------------------------------------------*/
/*
 * Check if the section is found, return 1 on success, 0 on error
 */
static int CONFIG_get_section(char const* section, FILE* fp) {

	char read_line[CONFIG_MAX_CHAR];
	char trim_line[CONFIG_MAX_CHAR];		// trimed read_line
	char cmp_section[CONFIG_MAX_CHAR];		// section name to be compared

	// we assume the conf file is opened on entry
	
	// go to the beginning position
	fseek(fp, (long)0, SEEK_SET);

	memset(read_line, 0, CONFIG_MAX_CHAR);
	memset(trim_line, 0, CONFIG_MAX_CHAR);
	memset(cmp_section, 0, CONFIG_MAX_CHAR);
	sprintf(cmp_section, "[%s]", section);
	
	while (fgets(read_line, CONFIG_MAX_CHAR, fp)) {
		CONFIG_trim(read_line, trim_line);
		CONFIG_lowercase(trim_line);
		if (!strncmp(trim_line, cmp_section, CONFIG_MAX_CHAR)) {
			break;
		}
		memset(read_line, 0, CONFIG_MAX_CHAR);
		memset(trim_line, 0, CONFIG_MAX_CHAR);
	}

	if (feof(fp)) {
		return 0;
	}
	return 1;
}

/*
 * Get the next key/value pair (in lowercase) from the given fp.
 * Return 1 on success, 0 otherwise
 */
static int CONFIG_get_next_key_value(char* key, char* value, FILE* fp) {
	
	char read_line[CONFIG_MAX_CHAR];
	char read_key[CONFIG_MAX_CHAR];
	char read_value[CONFIG_MAX_CHAR];
	char trim_line[CONFIG_MAX_CHAR];
	char trim_key[CONFIG_MAX_CHAR];
	char trim_value[CONFIG_MAX_CHAR];
	char* ptr_equal;
	int ret = 1;

	memset(read_line, 0, CONFIG_MAX_CHAR);
	memset(trim_line, 0, CONFIG_MAX_CHAR);

	// try to get the next key/value	
	while (fgets(read_line, CONFIG_MAX_CHAR, fp)) {
	
		CONFIG_trim(read_line, trim_line);
		if (trim_line[0] != '\0' && trim_line[0] != '#' && 
				trim_line[0] != '\r' && trim_line[0] != '\n' && 
				trim_line[0] != '[') {	
		
			if ((ptr_equal = strchr(read_line, '=')) == NULL) {
				debug_err("The line `%s' does not have '=' sign.\n", read_line);
				ret = 0;	
				break;	
			}	
			memset(read_key, 0, CONFIG_MAX_CHAR);
			memset(read_value, 0, CONFIG_MAX_CHAR);
			memset(trim_key, 0, CONFIG_MAX_CHAR);
			memset(trim_value, 0, CONFIG_MAX_CHAR);
			strncpy(read_key, read_line, ptr_equal - read_line);
			strncpy(read_value, ptr_equal + 1, CONFIG_MAX_CHAR);
			CONFIG_trim(read_key, trim_key);
			CONFIG_lowercase(trim_key);
			CONFIG_trim(read_value, trim_value);
			CONFIG_lowercase(trim_value);
			break;

		} else if (read_line[0] == '[') {
			ret = 0;
			break;
		}
		memset(read_line, 0, CONFIG_MAX_CHAR);
		memset(trim_line, 0, CONFIG_MAX_CHAR);
	}

	// end of file
	if (feof(fp)) {
		ret = 0;
	}
	
	// copy the string to results
	if (ret) {
		strncpy(key, trim_key, CONFIG_MAX_CHAR);
		strncpy(value, trim_value, CONFIG_MAX_CHAR);
	}
	return ret;	
}


/*
 * Trim the leading / trailing ' ', '\t', '\r' and '\n'
 */
static void CONFIG_trim(char const* orig_str, char* ret_str) {
	
	int pos_start = 0, pos_end = 0;		// start / end position of the string
	int is_start_found = 0;		// is the starting point found

	int i;

	// check if orig_str == NULL
	if (orig_str == NULL) {
		debug_err("The string is NULL in CONFIG_trim\n");	
		exit(-1);
	}

	// start triming
	for (i=0; i < CONFIG_MAX_CHAR && orig_str[i] != '\0'; ++i) {
		if (orig_str[i] != ' ' && orig_str[i] != '\t' && 
				orig_str[i] != '\r' && orig_str[i] != '\n') {
			if (!is_start_found) {	
				is_start_found = 1;
				pos_start = i;
			}
			pos_end = i + 1;		
		}	
	}
	if (i == CONFIG_MAX_CHAR) {
		debug_err("The line `%s' is too long.\n", orig_str);
		exit(-1);
	}

	// return the string
	if (pos_end == 0) {
		ret_str[0] = '\0';
	} else {
		strncpy(ret_str, orig_str + pos_start, pos_end - pos_start);
	}	
}

/*
 * convert the string to the lowercase
 */
static void CONFIG_lowercase(char* str) {
	int i;

	// check if orig_str == NULL
	if (str == NULL) {
		debug_err("The string is NULL in CONFIG_lowercase\n");
		exit(-1);
	}

	// convert to lowercase
	for (i=0; i < CONFIG_MAX_CHAR && str[i] != '\0'; ++i) {
		str[i] = tolower(str[i]);
	}
	if (i == CONFIG_MAX_CHAR) {
		debug_err("The line `%s' is too long.\n", str);
		exit(-1);
	}
}

