/*
 * - message_queue.c
 *
 * holding the messages that are not yet decrypted
 */

#include "seal.h"

/*
 * init the message queue
 */
MSGQUEUE_control_p_t MSGQUEUE_init() {
	MSGQUEUE_control_p_t msg_queue;

	msg_queue = (MSGQUEUE_control_p_t)malloc(sizeof(MSGQUEUE_control_t));

	msg_queue->head = NULL;
	msg_queue->tail = NULL;
	msg_queue->queue_size = 0;
	
	pthread_mutex_init(&(msg_queue->lock), NULL);
	pthread_cond_init(&(msg_queue->cond), NULL);

	return msg_queue;
}

/*-------------------------------------------------------------------------
 *						QUEUE MANAGEMENT
 *-------------------------------------------------------------------------*/
/*
 * Enqueue a message.
 */
void MSGQUEUE_enq(MSGQUEUE_control_p_t msg_queue, int rekey_seq, 
		char const* msg_sender, void const* msg, int msg_size) {

	MSGQUEUE_node_p_t new_node_p;

	// malloc the new node
	new_node_p = (MSGQUEUE_node_p_t)malloc(sizeof(MSGQUEUE_node_t));
	new_node_p->rekey_seq = rekey_seq;
	memcpy(new_node_p->msg_sender, msg_sender, SEAL_GUID_LEN);
	new_node_p->msg = malloc(msg_size);
	memcpy(new_node_p->msg, msg, msg_size);
	new_node_p->msg_size = msg_size;
	new_node_p->prev = NULL;	
	new_node_p->next = NULL;

	// add to the queue
	if (msg_queue->head == NULL) {
		msg_queue->head = new_node_p;
	} else {
		new_node_p->prev = msg_queue->tail;
		msg_queue->tail->next = new_node_p;
	}
	msg_queue->tail = new_node_p;

	// increment the size
	msg_queue->queue_size++;
}


/*
 * Dequeue a message from the queue
 * (assume the returned msg_sender and msg has allocated sufficient memory 
 * for holding the data)
 */
void MSGQUEUE_deq(MSGQUEUE_control_p_t msg_queue, char* msg_sender, 
		void* msg) {

	MSGQUEUE_node_p_t del_node_p;

	if (msg_queue->head == NULL) {
		debug_err("Dequeue from an empty message queue\n");
		return;
	}

	// get the head information
	del_node_p = msg_queue->head;
	memcpy(msg_sender, del_node_p->msg_sender, SEAL_GUID_LEN);
	memcpy(msg, del_node_p->msg, del_node_p->msg_size);

	// remove the head	
	if (msg_queue->head == msg_queue->tail) {
		msg_queue->head = NULL;
		msg_queue->tail = NULL;
	} else {
		msg_queue->head = del_node_p->next;
		msg_queue->head->prev = NULL;
	}

	// free resources
	free(del_node_p->msg);
	free(del_node_p);

	// decrement the size
	msg_queue->queue_size--;
}

/*
 * Return the rekey seq of the head entry of the queue, or -1 if head is null
 */
int MSGQUEUE_get_head_rekey_seq(MSGQUEUE_control_p_t msg_queue) {
	if (msg_queue->head == NULL) {
		return -1;
	}
	return msg_queue->head->rekey_seq;
}

/*
 * Return the msg size of the head entry of the queue, or -1 if head is null
 */
int MSGQUEUE_get_head_msg_size(MSGQUEUE_control_p_t msg_queue) {
	if (msg_queue->head == NULL) {
		return -1;
	}
	return msg_queue->head->msg_size;
}

/*
 * Return the queue size
 */
int MSGQUEUE_get_size(MSGQUEUE_control_p_t msg_queue) {
	return msg_queue->queue_size;
}

/*-------------------------------------------------------------------------
 *						DESTROY
 *-------------------------------------------------------------------------*/
/*
 * Destroy the message queue
 */
void MSGQUEUE_destroy(MSGQUEUE_control_p_t msg_queue) {
	MSGQUEUE_node_p_t del_node_p;

	while (msg_queue->head != NULL) {
		del_node_p = msg_queue->head;
		msg_queue->head = del_node_p->next;
		free(del_node_p->msg);
		free(del_node_p);
	}
	free(msg_queue);
}


