/*
 * receive_thread.c
 *
 * - (1) receiving any incoming packet, including data messages and membership
 *   messages, and (2) adding the message to the message_queue
 */

#include "seal.h"

static int mbox_g;
static pthread_t * process_thread_g;
static pthread_t * rekey_send_thread_g;
static pthread_t * rekey_poll_thread_g;
static PKTQUEUE_control_p_t pkt_queue_g;

void* RECEIVE_init(SEAL_session_p_t seal) {
	int data_len;						// data len return (-ve if errors)
	int service_type;					// service type (REG / MEMBERSHIP)
	char sender[SEAL_GUID_LEN];			// sender name	
	int no_guid;						// number of guids
	char guids[SEAL_MAX_GUID][SEAL_GUID_LEN];	// array of guid names
	int packet_type;					// message type (in packet.h)
	int endian_mismatch;				// check if endian matches
	char recv_buf[SEAL_MAX_PKT_SIZE];	// received data

	// set cancel state
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
	pthread_cleanup_push((void*)&RECEIVE_cleanup_handler, NULL);

	// set static variables
	mbox_g = seal->mbox;
	process_thread_g = seal->process_thread;	
	rekey_send_thread_g = seal->rekey_send_thread;
	rekey_poll_thread_g = seal->rekey_poll_thread;
	pkt_queue_g = seal->pkt_queue;

	// wait for an incoming packet
	while (1) {
		data_len = SP_receive(mbox_g, &service_type, sender, SEAL_MAX_GUID, 
				&no_guid, guids, (int16*)&packet_type, 
				&endian_mismatch, SEAL_MAX_PKT_SIZE, recv_buf);

		if (data_len >= 0) {
			// enqueue a message
			pthread_mutex_lock(&(pkt_queue_g->lock));
			PKTQUEUE_enq(pkt_queue_g, service_type, sender, no_guid, guids,
					packet_type, endian_mismatch, recv_buf, data_len);
			pthread_cond_broadcast(&(pkt_queue_g->cond));
			pthread_mutex_unlock(&(pkt_queue_g->lock));
		} else {
			// check for unexpected error
			SP_error(data_len);	

			// cancel all other threads
			if (*process_thread_g != 0) {
				pthread_cancel(*(seal->process_thread));
			}
			if (*rekey_poll_thread_g != 0) {
				pthread_cancel(*(seal->rekey_poll_thread));
			}
			if (*rekey_send_thread_g != 0) {
				pthread_cancel(*(seal->rekey_send_thread));
			}
		
			// quit this thread	
			return NULL;
		}

		// check the cancel point
		pthread_testcancel();	
	}       

	pthread_cleanup_pop(1);
	return 0;
}


void* RECEIVE_cleanup_handler() {
	pthread_mutex_trylock(&(pkt_queue_g->lock));
	pthread_mutex_unlock(&(pkt_queue_g->lock));
	debug("RECEIVE->cleanup\n");
	return NULL;
}
