/*
 * rekey_send_thread.c
 *
 * - retrieve rekey message from the queue and send out rekey messages
 */

#include "seal.h"

static KEYTREE_control_p_t keytree_engine_g;
static LEADER_control_p_t leader_engine_g;
static MEMBER_control_p_t member_engine_g;
static PACKET_control_p_t packet_engine_g;
static RQUEUE_control_p_t rekey_queue_g;
#ifdef SIGNATURE
static CERTKEY_control_p_t certkey_engine_g;
#endif

/*
 * send out rekey messages
 */
void* REKEY_SEND_init(SEAL_session_p_t seal) {
	char rekey_mode;
	char qsubtree_join[SEAL_GUID_LEN];		// used in QSUBTREE only
	char qsubtree_leave[SEAL_GUID_LEN];		// used in QSUBTREE only
	char zeros[SEAL_GUID_LEN];				// used in QSUBTREE only

	int no_mnode = -1;
	int no_join = -1;
	int no_leave = -1;
	char** mnode_guids = NULL;
	int* tree_node_ids = NULL;
	char** join_guids = NULL;
	char** leave_guids = NULL;
	KEYTREE_control_p_t tree;			// the tree to be managed
	int rekey_seq;						// rekey sequence number
	int is_rekey_needed = 1;			// check if rekey send is needed
	int i;

	PACKET_rekey_t pkt_rekey;			// the rekey packet to be sent
	int send_len;						// length of the sending data

	pthread_mutex_t dummy_lock;
	
	// set cancel state
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
	pthread_cleanup_push((void*)&REKEY_SEND_cleanup_handler, NULL);

	// set static variables
	keytree_engine_g = seal->keytree_engine;
	leader_engine_g = seal->leader_engine;
	member_engine_g = seal->member_engine;
	packet_engine_g = seal->packet_engine;
	rekey_queue_g = seal->rekey_queue;
#ifdef SIGNATURE
	certkey_engine_g = seal->certkey_engine;
#endif

	// init the lock
	pthread_mutex_init(&dummy_lock, NULL);

	while (1) {

		// dequeue a rekey message  
		pthread_mutex_lock(&(rekey_queue_g->lock));
		while (RQUEUE_get_stopflag(rekey_queue_g)) {
			pthread_cond_wait(&(rekey_queue_g->stop_cond),
					&(rekey_queue_g->lock));
		}
		while (RQUEUE_get_size(rekey_queue_g) <= 0) {
			pthread_cond_wait(&(rekey_queue_g->empty_cond),
					&(rekey_queue_g->lock));
		}
		RQUEUE_deq(rekey_queue_g, &rekey_mode, qsubtree_join, qsubtree_leave);
		RQUEUE_set_stopflag(rekey_queue_g, 1);
		pthread_mutex_unlock(&(rekey_queue_g->lock));

		// acquires all necessary locks
		pthread_mutex_lock(&(keytree_engine_g->lock));		
		pthread_mutex_lock(&(leader_engine_g->lock));
		pthread_mutex_lock(&(member_engine_g->lock));	
		pthread_mutex_lock(&(packet_engine_g->lock));
		
		// malloc and get the joins/leaves 
		if (rekey_mode != REKEY_QSUBTREE) {
			no_join = MEMBER_get_size(member_engine_g, MEMBER_JOIN_LIST);
			no_leave = MEMBER_get_size(member_engine_g, MEMBER_LEAVE_LIST);

			join_guids = (char**)malloc(no_join * sizeof(char*));
			for (i=0; i<no_join; ++i) {
				join_guids[i] = (char*)malloc(SEAL_GUID_LEN);
			}
			leave_guids = (char**)malloc(no_leave * sizeof(char*));
			for (i=0; i<no_leave; ++i) {
				leave_guids[i] = (char*)malloc(SEAL_GUID_LEN);
			}

			// get join/leave guids
			MEMBER_get_all(member_engine_g, MEMBER_JOIN_LIST, join_guids, NULL);
			MEMBER_get_all(member_engine_g, MEMBER_LEAVE_LIST, leave_guids, 
					NULL);

		} else {			// rekey_mode == REKEY_QSUBTREE
			
			// get join/leave guids
			memset(zeros, 0, SEAL_GUID_LEN);
			if (memcmp(qsubtree_join, zeros, SEAL_GUID_LEN) != 0) {
				no_join = 1;
				no_leave = 0;
				join_guids = (char**)malloc(no_join * sizeof(char*));
				join_guids[0] = (char*)malloc(SEAL_GUID_LEN);
				memcpy(join_guids[0], qsubtree_join, SEAL_GUID_LEN);
				leave_guids = NULL;		
			} else if (memcmp(qsubtree_leave, zeros, SEAL_GUID_LEN) != 0) {
				no_join = 0;
				no_leave = 1;
				join_guids = NULL;
				leave_guids = (char**)malloc(no_leave * sizeof(char*));
				leave_guids[0] = (char*)malloc(SEAL_GUID_LEN);
				memcpy(leave_guids[0], qsubtree_leave, SEAL_GUID_LEN);
			} else {
				debug_err("Wrong values on no_join/no_leave in qsubtree\n");
			}

			// check if rekey is necessary. It is not necessary if the
			// join/leave has been processed in the previous QMERGE message
			if (no_join > 0) {
				if (KEYTREE_is_guid_exist(keytree_engine_g, join_guids[0]) || 
						KEYTREE_is_guid_exist(keytree_engine_g->subtree, 
							join_guids[0])) {
					is_rekey_needed = 0;
				}
			} else if (no_leave > 0) {
				if (!KEYTREE_is_guid_exist(keytree_engine_g->subtree, 
							leave_guids[0])) {
					is_rekey_needed = 0;	
				}
			}		
		}			// end: rekey mode on getting joins/leaves

		// do only when rekey is needed. Not needed in qsubtree case
		if (is_rekey_needed) {		
			// get the keytree information
			if (rekey_mode != REKEY_QSUBTREE) {
				tree = keytree_engine_g;
			} else {
				tree = keytree_engine_g->subtree;
			}

			// malloc and get tree node data
			no_mnode = KEYTREE_get_num_of_members(tree);
			mnode_guids = (char**)malloc(no_mnode * sizeof(char*));
			for (i=0; i<no_mnode; ++i) {
				mnode_guids[i] = (char*)malloc(SEAL_GUID_LEN);
			}
			tree_node_ids = (int*)malloc(no_mnode * sizeof(int));

			// get tree node data   
			KEYTREE_get_members_tree_node_ids(tree, mnode_guids, 
					tree_node_ids);

			// get the sequence number
			LEADER_seq_increment(leader_engine_g);
			rekey_seq = LEADER_seq_retrieve(leader_engine_g);

			// bcast the rekey message
			memset(&pkt_rekey, 0, sizeof(PACKET_rekey_t));
			send_len = PACKET_create_rekey(rekey_seq, rekey_mode,no_mnode, 
					no_join, no_leave, mnode_guids, tree_node_ids, join_guids, 
					leave_guids, &pkt_rekey);
#ifdef SIGNATURE
			// sign the packet
			pthread_mutex_lock(&(certkey_engine_g->lock));
			if (CERTKEY_sign(certkey_engine_g, &pkt_rekey, send_len, 
						(char*)(&pkt_rekey) + send_len) == 0) {
				// attach the signature at the end of the packet    
				debug_err("Signature failed\n");
			}
			pthread_mutex_unlock(&(certkey_engine_g->lock));
			send_len += CERTKEY_LEN;
#endif
            PACKET_send(packet_engine_g, PACKET_REKEY, send_len, &pkt_rekey);
		}
		
		// releases all locks
		pthread_mutex_unlock(&(packet_engine_g->lock));
		pthread_mutex_unlock(&(member_engine_g->lock));	
		pthread_mutex_unlock(&(leader_engine_g->lock)); 
		pthread_mutex_unlock(&(keytree_engine_g->lock));      

		// free all resources
	
		for (i=0; i<no_join; ++i)
			free(join_guids[i]);
		for (i=0; i<no_leave; ++i)
			free(leave_guids[i]);
		if (join_guids != NULL) {
			free(join_guids);
		}
		if (leave_guids != NULL) {
			free(leave_guids);
		}
		if (is_rekey_needed) {
			for (i=0; i<no_mnode; ++i)
				free(mnode_guids[i]);
			free(mnode_guids);
			free(tree_node_ids);
		}

	}       // end: while(1) loop

	pthread_cleanup_pop(1);

	return NULL;
}


/*
 * cleanup handler
 */
void* REKEY_SEND_cleanup_handler() {
	pthread_mutex_trylock(&(keytree_engine_g->lock));       
	pthread_mutex_trylock(&(leader_engine_g->lock)); 
	pthread_mutex_trylock(&(member_engine_g->lock));      
	pthread_mutex_trylock(&(packet_engine_g->lock));
	pthread_mutex_trylock(&(rekey_queue_g->lock));
	
	pthread_mutex_unlock(&(rekey_queue_g->lock));
	pthread_mutex_unlock(&(packet_engine_g->lock));
	pthread_mutex_unlock(&(member_engine_g->lock));      
	pthread_mutex_unlock(&(leader_engine_g->lock)); 
	pthread_mutex_unlock(&(keytree_engine_g->lock));       

	debug("REKEY send->cleanup\n");
	return NULL;
}

