/*
 * sesskey_engine.h
 */

#define SESSKEY_LEN					128		// in number of bytes
#define SESSKEY_FILE				"./dh1024.pem"	// parameters

/* Parameters in DH */
typedef struct sesskey_param_s * SESSKEY_param_p_t;
typedef struct sesskey_param_s {
	char sess_sec_key[SESSKEY_LEN];			// session secret key
	char sess_bli_key[SESSKEY_LEN];			// session blinded key
	char g[SESSKEY_LEN];						// geneartor
	char p[SESSKEY_LEN];						// prime p
} SESSKEY_param_t;

/* Key path holding the keys */
typedef struct sesskey_keypath_entry_s * SESSKEY_keypath_entry_p_t;
typedef struct sesskey_keypath_entry_s {
	int tree_node_id;				// own's tree node id
	int sib_tree_node_id;			// sibling's tree node id
	char sec_key[SESSKEY_LEN];		// own's secret key
	char sib_bli_key[SESSKEY_LEN];	// sibling's blinded key
	char bcast_flag;				// check if bcast is required
	char sec_key_flag;				// check if sec_key needs update
	char bli_key_flag;				// check if bli_key needs update
	SESSKEY_keypath_entry_p_t prev;		
	SESSKEY_keypath_entry_p_t next;	
} SESSKEY_keypath_entry_t;

typedef struct sesskey_keypath_s * SESSKEY_keypath_p_t;
typedef struct sesskey_keypath_s {
	int rekey_seq;		// seq # of computing the pri key w.r.t. rekey msg
	SESSKEY_keypath_entry_p_t head;
	SESSKEY_keypath_entry_p_t tail;
} SESSKEY_keypath_t;

/* Cache temporary holding the pub key received */
typedef struct sesskey_cache_entry_s * SESSKEY_cache_entry_p_t;
typedef struct sesskey_cache_entry_s {
	int tree_node_id;					// tree node id
	char bli_key[SESSKEY_LEN];			// bli_key	
	int rekey_seq;						// last modified timestamp
	SESSKEY_cache_entry_p_t prev;
	SESSKEY_cache_entry_p_t next;
} SESSKEY_cache_entry_t;

typedef struct sesskey_cache_s * SESSKEY_cache_p_t;
typedef struct sesskey_cache_s {
	SESSKEY_cache_entry_p_t head;
	SESSKEY_cache_entry_p_t tail;
} SESSKEY_cache_t;

/* A list of group keys */
typedef struct sesskey_groupkey_entry_s * SESSKEY_groupkey_entry_p_t;
typedef struct sesskey_groupkey_entry_s {
	char group_key[SESSKEY_LEN];			// group_key	
	int rekey_seq;						// rekey sequence number
	SESSKEY_groupkey_entry_p_t prev;
	SESSKEY_groupkey_entry_p_t next;
} SESSKEY_groupkey_entry_t;

typedef struct sesskey_groupkey_s * SESSKEY_groupkey_p_t;
typedef struct sesskey_groupkey_s {
	int size;
	SESSKEY_groupkey_entry_p_t head;
	SESSKEY_groupkey_entry_p_t tail;
} SESSKEY_groupkey_t;

/* Key parameters / individual keys held by the servent */
typedef struct sesskey_control_s * SESSKEY_control_p_t;
typedef struct sesskey_control_s {
	SESSKEY_param_p_t param;						// the parameters structure
	SESSKEY_keypath_p_t keypath;
	SESSKEY_cache_p_t cache;					
	SESSKEY_groupkey_p_t groupkeys;
	pthread_mutex_t lock;
} SESSKEY_control_t;


/************ function prototypes *********************/

/* init the key engine */
SESSKEY_control_p_t SESSKEY_init();

/* keypath management */
SESSKEY_keypath_p_t SESSKEY_keypath_create(int len);
int SESSKEY_keypath_update(SESSKEY_control_p_t key_engine, int rekey_seq,
		int const* new_keypath_tree_node_ids, int new_keypath_len, 
		int const* renewed_list, int no_renewed, int const* bcast_list, 
		int no_bcast);
int SESSKEY_keypath_compute(SESSKEY_control_p_t key_engine, 
		int sib_tree_node_id, char const* bli_key, int seq);
int SESSKEY_keypath_retrieve_bcast(SESSKEY_control_p_t key_engine, 
		int start_tree_node_id, int* ret_tree_node_ids, 
		char** ret_bcast_keys);
void SESSKEY_keypath_unmark_bcast(SESSKEY_control_p_t key_engine, 
		int tree_node_id);
void SESSKEY_keypath_to_groupkey(SESSKEY_control_p_t key_engine, int seq);
int SESSKEY_keypath_get_len(SESSKEY_control_p_t key_engine);
int SESSKEY_keypath_is_root(SESSKEY_control_p_t key_engine);
void SESSKEY_keypath_print(SESSKEY_control_p_t key_engine);

/* cache management */
void SESSKEY_cache_insert(SESSKEY_control_p_t key_engine, int tree_node_id, 
		char const* bli_key, int seq);
void SESSKEY_cache_delete_expire(SESSKEY_control_p_t key_engine, int rekey_seq);
int SESSKEY_cache_update(SESSKEY_control_p_t key_engine, int tree_node_id, 
		char const* bli_key, int seq);
int SESSKEY_cache_retrieve(SESSKEY_control_p_t key_engine, char* bli_key, 
		int tree_node_id, int seq);
void SESSKEY_cache_to_keypath(SESSKEY_control_p_t key_engine);
int SESSKEY_cache_is_exist(SESSKEY_control_p_t key_engine, int tree_node_id);
void SESSKEY_cache_print(SESSKEY_control_p_t key_engine);

/* groupkey management */
void SESSKEY_groupkey_insert(SESSKEY_control_p_t key_engine, 
		char const* group_key, int seq);
int SESSKEY_groupkey_retrieve(SESSKEY_control_p_t key_engine, char* group_key, 
		int seq);
int SESSKEY_groupkey_is_empty(SESSKEY_control_p_t key_engine);
int SESSKEY_groupkey_is_formed(SESSKEY_control_p_t key_engine, int seq);
void SESSKEY_groupkey_print(SESSKEY_control_p_t key_engine);

/* key computation */
void SESSKEY_compute_secret(SESSKEY_control_p_t key_engine, char* parent_key, 
		char const* sec_key, char const* bli_key);
void SESSKEY_compute_blinded(SESSKEY_control_p_t key_engine, char* blinded_key,
		char const* sec_key);

/* encryption / decryption */
int SESSKEY_encrypt(SESSKEY_control_p_t key_engine, char const* in_data, 
		char* out_data, int len);
int SESSKEY_decrypt(SESSKEY_control_p_t key_engine, char const* in_data, 
		char* out_data, int len, int seq);

/* key confirmation */
int SESSKEY_confirm(SESSKEY_control_p_t key_engine, int rekey_seq);

/* destory */
void SESSKEY_destroy(SESSKEY_control_p_t key_engine);
