/*
 * ui_engine.c
 *
 * - providing a graphical user input interface for the chatter
 * - using gtk for development
 * - supporting both GUI and textmode
 */

#include "chatter.h"

/*-----------------------------------------------------------------------
 *						GUI and Textmode Functions
 *----------------------------------------------------------------------*/
/*
 * Intialize the ui engine
 */
UI_control_p_t UI_init(int argc, char** argv) {

	UI_control_p_t ui_engine;

	// malloc
	ui_engine = (UI_control_p_t)malloc(sizeof(UI_control_t));

#ifdef GUI_ENABLED					// guid mode

	// initialize the gtk toolkit 
	gtk_init(&argc, &argv);

	// init main window
	UI_init_gui_main(ui_engine);
#ifdef SIGNATURE				// signature mode
	UI_init_gui_passwd(ui_engine);
#endif							// end: signature mode
#endif								// end: gui mode

	ui_engine->state = UI_GROUP_OUT;		// outside the group	

	return ui_engine;
}

/*
 * Start accepting input (the application should block here)
 */
void UI_start(UI_control_p_t ui_engine) {

#ifdef GUI_ENABLED			// gui mode

#ifdef SIGNATURE	// signature mode
	// show interface that requests passwd	
	gtk_widget_show_all(ui_engine->passwd_window);
#else				// non-signature mode
	// show all components
	gtk_widget_show_all(ui_engine->main_window);
#endif				// end: signature mode
	
	// wait for inputs	
	gtk_main();

#else						// textmode

	char* input;
	char** arg;
	int no_arg = UI_MAX_ARG;
	int i;

#ifdef SIGNATURE	// signature mode
	UI_textmode_passwd_check();
#endif				// end: signature mode

	// init the variable
	input = (char*)malloc(UI_MAX_CHAR * UI_MAX_ARG);
	arg = (char**)malloc(UI_MAX_ARG * sizeof(char*));
	for (i=0; i<UI_MAX_ARG; ++i) {
		arg[i] = (char*)malloc(UI_MAX_CHAR);
	}

	// print some basic information
	printf("\n");
	printf("UID: %s\n", seal_session_g->uid);
	printf("SPREAD IP: %s\n", seal_session_g->spread_ip);
	printf("SPREAD PORT: %d\n", seal_session_g->spread_port);

	// print the prompt
	while (1) {
		memset(input, 0, UI_MAX_CHAR);
		for (i=0; i<no_arg; ++i) {
			memset(arg[i], 0, UI_MAX_CHAR);
		}

		printf("CHATTER>");
		fgets(input, UI_MAX_CHAR, stdin);
		no_arg = UI_textmode_split_arg(input, arg);

		if (strncmp(arg[0], "help", UI_MAX_CHAR) == 0) {
			printf("Commands\n");
			printf("\thelp -- display the help menu\n");	
			printf("\tjoin -- join the group\n");
			printf("\tleave -- leave the group\n");
			printf("\tmember -- print the current membership status\n");
			printf("\tmsg <msg data> -- send the message\n");
			printf("\texit -- exit the application\n");
			fflush(stdout);

		} else if (strncmp(arg[0], "join", UI_MAX_CHAR) == 0) {
			UI_textmode_join(ui_engine, no_arg, arg);

		} else if (strncmp(arg[0], "leave", UI_MAX_CHAR) == 0) {
			UI_textmode_leave(ui_engine, no_arg, arg);

		} else if (strncmp(arg[0], "member", UI_MAX_CHAR) == 0) {
			UI_textmode_member(ui_engine, no_arg, arg);

		} else if (strncmp(arg[0], "msg", UI_MAX_CHAR) == 0) {
			UI_textmode_msg(ui_engine, input + 4);

		} else if (strncmp(arg[0], "exit", UI_MAX_CHAR) == 0) {
			UI_textmode_exit(no_arg, arg);
		
		} 
	}
#endif
}

/*
 * Print the members which are currently in the group
 * (the guid is in the `#uid#group' format)
 */
void UI_print_members(UI_control_p_t ui_engine, char* const* members,
		int no_member) {

	char* curr_uid[1];				// gtk list functions need this declaration
	int i;

#ifdef GUI_ENABLED					// GUI mode
	// malloc
	curr_uid[0] = (char*)malloc(SEAL_UID_LEN);

	// clear the list
	gtk_clist_clear(GTK_CLIST(ui_engine->member_clist));

	// add members to the member_clist
	for (i=0; i<no_member; ++i) {
		// retrieve the curr uid
		memset(curr_uid[0], 0, SEAL_UID_LEN);
		if (guid_to_uid(members[i], curr_uid[0]) == 0) {
			fprintf(stderr, "ERROR: Incorrect GUID in curr member\n");
			return;
		}
		gtk_clist_append(GTK_CLIST(ui_engine->member_clist), curr_uid);
	}

	// free the malloc resource
	free(curr_uid[0]);

#else								// textmode
	// malloc
	curr_uid[0] = (char*)malloc(SEAL_UID_LEN);

	printf("===============================================\n");
	printf("Current membership information:\n\n");
	for (i=0; i<no_member; ++i) {
		// retrieve the curr uid
		memset(curr_uid[0], 0, SEAL_UID_LEN);
		if (guid_to_uid(members[i], curr_uid[0]) == 0) {
			fprintf(stderr, "ERROR: Incorrect GUID in curr member\n");
			return;
		}
		printf("%5d\t%s\n", i+1, curr_uid[0]);
	}	
	printf("\n===============================================\n");

	// free the malloc resource
	free(curr_uid[0]);	
#endif
	
}

/*
 * Print the message received
 */
void UI_print_msg(UI_control_p_t ui_engine, char const* sender, 
		char const* msg) {
	char sender_uid[SEAL_UID_LEN];	
	char* print_line[1];
	time_t now;
	char now_str[10];

	// malloc
	print_line[0] = (char*)malloc(strlen(msg) + 50);
	
	// retrieve the sender's uid
	if (guid_to_uid(sender, sender_uid) == 0) {
		fprintf(stderr, "ERROR: Incorrect UID in member in UI_print_msg\n");
		return;
	}	

	// get the current time
	now = time(NULL);
	memset(now_str, 0, sizeof(now_str));
	strftime(now_str, sizeof(now_str), "%H:%M", localtime(&now));

	// prepare the print line
	sprintf(print_line[0], "(%s at %s): %s", sender_uid, now_str, msg);

#ifdef GUI_ENABLED					// GUI mode
	gtk_clist_prepend(GTK_CLIST(ui_engine->msg_clist), print_line);
		
#else								// text mode
	printf("Message received %s\n", print_line[0]);
	fflush(stdout);
#endif

	free(print_line[0]);
}

/*
 * Print the error messages
 */
void UI_print_error(UI_control_p_t ui_engine, int err_code, char* arg) {
	char* print_line[1];

	// malloc
	print_line[0] = (char*)malloc(UI_MAX_PRINT_LEN);

	// determine the error messages
	switch (err_code) {
		case UI_ERR_GROUP_NOT_VALID:
			sprintf(print_line[0], 
					"Error: The group to be joined is not valid.");
			break;
		case UI_ERR_GROUP_JOINED:
			sprintf(print_line[0], 
					"Error: You are already in the group.");
			break;
		case UI_ERR_GROUP_LEFT:
			sprintf(print_line[0], 
					"Error: You are not in any group.");
			break;
		case UI_ERR_GROUP_NO_MEMBER:
			sprintf(print_line[0], 
					"Error: No member is defined in the group.");
			break;
		case UI_ERR_MSG_NOT_ENCRYPTED:
			sprintf(print_line[0], 
					"Error: Message is not sent since it is not encrypted.");
			break;
		case UI_ERR_MEMBER_NOT_AUTHORIZED:
			sprintf(print_line[0],
					"Error:	An unauthorized member joins the group.");
		default:
			break;
	}

	if (arg != NULL) {
		strncat(print_line[0], " : ", UI_MAX_PRINT_LEN);
		strncat(print_line[0], arg, UI_MAX_PRINT_LEN);	
	}	


#ifdef GUI_ENABLED					// GUI mode
	gtk_clist_prepend(GTK_CLIST(ui_engine->status_clist), print_line);
#else								// text mode
	printf("%s\n", print_line[0]);
	fflush(stdout);
#endif
	
	free(print_line[0]);
}


/*
 * Wait for incoming messages
 */
void* UI_receive_handler(UI_control_p_t ui_engine) {
	char sender[SEAL_GUID_LEN];
	char msg[SEAL_MAX_MSG_SIZE];

	// set cancel state
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
			
	// loop to wait for messages    
	while (1) {
		memset(sender, 0, SEAL_GUID_LEN);
		memset(msg, 0, SEAL_MAX_MSG_SIZE);
		if (SEAL_recv(seal_session_g, sender, msg) >= 0) {
			UI_print_msg(ui_engine, sender, msg);
		} else {
			fprintf(stderr, "ERROR: error occurs in msg recv\n");
			break;
		}
	}
	return NULL;
}


/*-----------------------------------------------------------------------
 *							GUI FUNCTIONS
 *-----------------------------------------------------------------------*/

#ifdef GUI_ENABLED

/*-----------------------------------------------------------------------
 *					GUI: Initialization functions
 *----------------------------------------------------------------------*/
/*
 * init the main window of the GUI
 * (simply initialize the components but do not show them in this function)
 */
void UI_init_gui_main(UI_control_p_t ui_engine) {
	char main_window_title[100];

	// main window: create (set the title with the uid)
	// vbox --> table --> box
	memset(main_window_title, 0, sizeof(main_window_title));
	sprintf(main_window_title, "Chatter (%s)", seal_session_g->uid);
	ui_engine->main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(ui_engine->main_window), 
			main_window_title);
//	gtk_widget_set_size_request(GTK_WIDGET(ui_engine->main_window), 650, 535);
	gtk_container_set_border_width(GTK_CONTAINER(ui_engine->main_window), 5);
	g_signal_connect(G_OBJECT(ui_engine->main_window), "delete_event",
			G_CALLBACK(UI_gui_exit), NULL);
	g_signal_connect(G_OBJECT(ui_engine->main_window), "destroy",
			G_CALLBACK(UI_gui_exit), NULL);

	ui_engine->main_vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(ui_engine->main_vbox), 5);
	gtk_container_add(GTK_CONTAINER(ui_engine->main_window), 
			ui_engine->main_vbox);

	ui_engine->main_table = gtk_table_new(3, 2, FALSE);	
	gtk_table_set_row_spacings(GTK_TABLE(ui_engine->main_table), 5);
	gtk_table_set_col_spacings(GTK_TABLE(ui_engine->main_table), 5);
	gtk_container_set_border_width(GTK_CONTAINER(ui_engine->main_table), 5);
	gtk_box_pack_start(GTK_BOX(ui_engine->main_vbox), ui_engine->main_table, 
			TRUE, TRUE, 0);	
	
	ui_engine->main_button_hbox = gtk_hbox_new(FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(ui_engine->main_button_hbox),
			5);
	gtk_widget_set_size_request(GTK_WIDGET(ui_engine->main_button_hbox), -1,
			35);
	gtk_box_pack_start(GTK_BOX(ui_engine->main_vbox),
			ui_engine->main_button_hbox, TRUE, TRUE, 0);

	// input frame: create and pack
	// hbox --> label --> text_entry --> button
	ui_engine->input_frame = gtk_frame_new("Input");
	gtk_widget_set_size_request(GTK_WIDGET(ui_engine->input_frame), 640, 55);
	gtk_table_attach_defaults(GTK_TABLE(ui_engine->main_table), 
			ui_engine->input_frame, 0, 2, 0, 1);

	ui_engine->input_hbox = gtk_hbox_new(FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(ui_engine->input_hbox), 5);
	gtk_container_add(GTK_CONTAINER(ui_engine->input_frame),
			ui_engine->input_hbox);

	ui_engine->input_label = gtk_label_new("Input Message: ");
	gtk_box_pack_start(GTK_BOX(ui_engine->input_hbox), 
			ui_engine->input_label, TRUE, TRUE, 5);	

	ui_engine->input_text_entry = gtk_entry_new();
	gtk_widget_set_size_request(ui_engine->input_text_entry, 400, -1);
	gtk_box_pack_start(GTK_BOX(ui_engine->input_hbox), 
			ui_engine->input_text_entry, TRUE, TRUE, 5);	

	ui_engine->input_button = gtk_button_new_with_label("Send");
	//gtk_widget_set_size_request(ui_engine->input_button, 70, -1);
	gtk_box_pack_start(GTK_BOX(ui_engine->input_hbox), 
			ui_engine->input_button, TRUE, TRUE, 5);	
	g_signal_connect(G_OBJECT(ui_engine->input_button),
			"clicked", G_CALLBACK(UI_gui_input_send), ui_engine);

	// message frame: create and pack
	// vbox --> scrolled_window --> list --> clear_button
	ui_engine->msg_frame = gtk_frame_new("Message");
	gtk_widget_set_size_request(GTK_WIDGET(ui_engine->msg_frame), 400, 320);
	gtk_table_attach_defaults(GTK_TABLE(ui_engine->main_table), 
			ui_engine->msg_frame, 0, 1, 1, 2);

	ui_engine->msg_vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(ui_engine->msg_vbox), 5);
	gtk_container_add(GTK_CONTAINER(ui_engine->msg_frame), 
			ui_engine->msg_vbox);
	
	ui_engine->msg_scrolled_window = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(
				ui_engine->msg_scrolled_window),
			GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);	
	gtk_box_pack_start(GTK_BOX(ui_engine->msg_vbox),
			ui_engine->msg_scrolled_window, TRUE, TRUE, 5);	

	ui_engine->msg_clist = gtk_clist_new(1);
	gtk_clist_set_column_justification(GTK_CLIST(ui_engine->msg_clist), 0, 
			GTK_JUSTIFY_LEFT);
	gtk_widget_set_size_request(ui_engine->msg_clist, -1, 250);
	gtk_container_add(GTK_CONTAINER(ui_engine->msg_scrolled_window), 
			ui_engine->msg_clist);

	ui_engine->msg_clear_button = gtk_button_new_with_label("Clear");
	gtk_box_pack_start(GTK_BOX(ui_engine->msg_vbox), 
			ui_engine->msg_clear_button, TRUE, TRUE, 5);	
	g_signal_connect(G_OBJECT(ui_engine->msg_clear_button), 
			"clicked", G_CALLBACK(UI_gui_msg_clear), ui_engine);
	
	// membership frame: create and pack
	// member table --> scrolled_window --> list
	// --> group label --> group entry --> join button --> leave button
	ui_engine->member_frame = gtk_frame_new("Membership");
	gtk_widget_set_size_request(GTK_WIDGET(ui_engine->member_frame), 210, 320);
	gtk_table_attach_defaults(GTK_TABLE(ui_engine->main_table), 
			ui_engine->member_frame, 1, 2, 1, 2);

	ui_engine->member_table = gtk_table_new(4, 2, FALSE);	
	gtk_table_set_row_spacings(GTK_TABLE(ui_engine->member_table), 10);
	gtk_table_set_col_spacings(GTK_TABLE(ui_engine->member_table), 10);
	gtk_container_set_border_width(GTK_CONTAINER(ui_engine->member_table), 10);
	gtk_container_add(GTK_CONTAINER(ui_engine->member_frame), 
			ui_engine->member_table);
	
	ui_engine->member_scrolled_window = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(
				ui_engine->member_scrolled_window),
			GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);	
	gtk_table_attach_defaults(GTK_TABLE(ui_engine->member_table), 
			ui_engine->member_scrolled_window, 0, 2, 0, 1);

	ui_engine->member_clist = gtk_clist_new(1);
	gtk_clist_set_column_justification(GTK_CLIST(ui_engine->member_clist), 0, 
			GTK_JUSTIFY_LEFT);
	gtk_widget_set_size_request(ui_engine->member_clist, -1, 180);
	gtk_container_add(GTK_CONTAINER(ui_engine->member_scrolled_window), 
			ui_engine->member_clist);

	ui_engine->member_group_label = gtk_label_new("Enter Your Group Below:");
	gtk_table_attach_defaults(GTK_TABLE(ui_engine->member_table), 
			ui_engine->member_group_label, 0, 2, 1, 2);
	
	ui_engine->member_group_entry = gtk_entry_new();
	gtk_table_attach_defaults(GTK_TABLE(ui_engine->member_table), 
			ui_engine->member_group_entry, 0, 2, 2, 3);

	ui_engine->member_join_button = gtk_button_new_with_label("Join");
	gtk_widget_set_size_request(GTK_WIDGET(
				ui_engine->member_join_button), 90, -1);
	gtk_table_attach_defaults(GTK_TABLE(ui_engine->member_table), 
			ui_engine->member_join_button, 0, 1, 3, 4);
	g_signal_connect(G_OBJECT(ui_engine->member_join_button), 
			"clicked", G_CALLBACK(UI_gui_join), ui_engine);

	ui_engine->member_leave_button = gtk_button_new_with_label("Leave");
	gtk_widget_set_size_request(GTK_WIDGET(
				ui_engine->member_leave_button), 90, -1);
	gtk_table_attach_defaults(GTK_TABLE(ui_engine->member_table), 
			ui_engine->member_leave_button, 1, 2, 3, 4);
	g_signal_connect(G_OBJECT(ui_engine->member_leave_button), 
			"clicked", G_CALLBACK(UI_gui_leave), ui_engine);

	// status frame: create and pack
	// scrolled_window --> clist
	ui_engine->status_frame = gtk_frame_new("Status");
	gtk_widget_set_size_request(ui_engine->status_frame, -1, 80);
	gtk_container_set_border_width(GTK_CONTAINER(ui_engine->status_frame), 5);
	gtk_table_attach_defaults(GTK_TABLE(ui_engine->main_table), 
			ui_engine->status_frame, 0, 2, 2, 3);

	ui_engine->status_scrolled_window = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(
				ui_engine->status_scrolled_window),
			GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);	
	gtk_container_add(GTK_CONTAINER(ui_engine->status_frame), 
			ui_engine->status_scrolled_window);

	ui_engine->status_clist = gtk_clist_new(1);
	gtk_container_add(GTK_CONTAINER(ui_engine->status_scrolled_window), 
			ui_engine->status_clist);

	// buttons: add to ui_engine->main_button_hbox
	// exit
	ui_engine->button_exit = gtk_button_new_with_label("Exit Application");
	gtk_box_pack_start(GTK_BOX(ui_engine->main_button_hbox), 
			ui_engine->button_exit, TRUE, TRUE, 150);
	g_signal_connect(G_OBJECT(ui_engine->button_exit), "clicked",
			G_CALLBACK(UI_gui_exit), NULL);
}


#ifdef SIGNATURE
/*
 * Init the passwd window of the GUI
 */
void UI_init_gui_passwd(UI_control_p_t ui_engine) {

	ui_engine->passwd_window = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(ui_engine->passwd_window), "Password");
	g_signal_connect(G_OBJECT(ui_engine->passwd_window), "delete_event",
			G_CALLBACK(UI_gui_exit), NULL);

	// vbox
	// label --> entry
	ui_engine->passwd_label = gtk_label_new("Enter your password");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(ui_engine->passwd_window)->vbox),
			ui_engine->passwd_label, TRUE, TRUE, 5);

	ui_engine->passwd_entry = gtk_entry_new();
	gtk_entry_set_visibility(GTK_ENTRY(ui_engine->passwd_entry), FALSE);
	gtk_entry_set_max_length(GTK_ENTRY(ui_engine->passwd_entry), 
			SEAL_PASSWD_LEN);
	gtk_widget_set_size_request(ui_engine->passwd_entry, 150, -1);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(ui_engine->passwd_window)->vbox),
			ui_engine->passwd_entry, TRUE, TRUE, 5);

	// action area
	// ok --> cancel
	ui_engine->passwd_button_ok = gtk_button_new_with_label("Ok");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(
					ui_engine->passwd_window)->action_area), 
			ui_engine->passwd_button_ok, TRUE, TRUE, 5);
	g_signal_connect(G_OBJECT(ui_engine->passwd_button_ok), "clicked", 
			G_CALLBACK(UI_gui_passwd_check), ui_engine);
	
	ui_engine->passwd_button_cancel = gtk_button_new_with_label("Cancel");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(
					ui_engine->passwd_window)->action_area),
			ui_engine->passwd_button_cancel, TRUE, TRUE, 5);
	g_signal_connect(G_OBJECT(ui_engine->passwd_button_cancel), "clicked", 
			G_CALLBACK(UI_gui_exit), NULL);
}
#endif


/*-----------------------------------------------------------------------
 *					GUI: Callback functions
 *----------------------------------------------------------------------*/

#ifdef SIGNATURE		

/*---------------------------
 * Signature mode for GUI
 *---------------------------*/
/*
 * Callback function: check the password. If valid, show the main window, 
 */
void UI_gui_passwd_check(GtkWidget* widget, UI_control_p_t ui_engine) {
	char password[SEAL_PASSWD_LEN + 1];		// extra byte for null character

	// get the password phrase 
	memset(password, 0, SEAL_PASSWD_LEN + 1);
	strncpy(password, gtk_entry_get_text(GTK_ENTRY(ui_engine->passwd_entry)),
			SEAL_PASSWD_LEN);

	// check against password
	if (SEAL_set_passwd(seal_session_g, password)) {
		gtk_widget_hide(ui_engine->passwd_window);
		gtk_widget_show_all(ui_engine->main_window);
	} else {
		gtk_label_set_text(GTK_LABEL(ui_engine->passwd_label), 
				"Invalid password. Enter again!");
		gtk_entry_set_text(GTK_ENTRY(ui_engine->passwd_entry), "");
	}
}

#endif	// end: ifdef signature flag

/*
 * Callback function: input_frame -- send message
 */
void UI_gui_input_send(GtkWidget* widget, UI_control_p_t ui_engine) {
	char plain_msg[UI_MAX_MSG_SIZE];

	if (ui_engine->state == UI_GROUP_IN) {
		// prepare the message
		memset(plain_msg, 0, UI_MAX_MSG_SIZE);
		strncpy(plain_msg, 
				gtk_entry_get_text(GTK_ENTRY(ui_engine->input_text_entry)), 
				UI_MAX_MSG_SIZE);

		// send message
		if (SEAL_send(seal_session_g, plain_msg, strlen(plain_msg)) < 0) {
			UI_print_error(ui_engine, UI_ERR_MSG_NOT_ENCRYPTED, NULL);
		}

		// clear the text entry field
		gtk_entry_set_text(GTK_ENTRY(ui_engine->input_text_entry), "");
	}  else {
		UI_print_error(ui_engine, UI_ERR_GROUP_LEFT, seal_session_g->group);
	}
}

/*
 * Callback function: message_frame - clear clist data
 */
void UI_gui_msg_clear(GtkWidget* widget, UI_control_p_t ui_engine) {
	gtk_clist_clear(GTK_CLIST(ui_engine->msg_clist));
}


/*
 * Callback function: member_frame -- handle the join event
 */
void UI_gui_join(GtkWidget* widget, UI_control_p_t ui_engine) {

	char group[SEAL_GROUP_LEN];
	char* print_line[] = {"Status: You have successfully joined a group."};

	if (ui_engine->state == UI_GROUP_OUT) {

		// check if group name is inputted
		if (strlen(gtk_entry_get_text(GTK_ENTRY(
							ui_engine->member_group_entry))) > 0) {
			
			// set the group name
			memset(group, 0, SEAL_GROUP_LEN);
			strncpy(group, gtk_entry_get_text(GTK_ENTRY(
							ui_engine->member_group_entry)), SEAL_GROUP_LEN);

			// join the group
			if (SEAL_join(seal_session_g, group)) {	
				// succeed in joining the group
				gtk_entry_set_editable(GTK_ENTRY(
							ui_engine->member_group_entry), FALSE);	
				ui_engine->state = UI_GROUP_IN;

				// start the threads
				pthread_create(&(ui_engine->poll_member_thread), NULL, 
							(void*)&UI_gui_poll_members, ui_engine);
				pthread_create(&(ui_engine->msg_recv_thread), NULL, 
					(void*)&UI_receive_handler, ui_engine);

				// report status
				gtk_clist_prepend(GTK_CLIST(ui_engine->status_clist), 
						print_line);

			} else {						// error in joining the group
				UI_print_error(ui_engine, UI_ERR_GROUP_NOT_VALID, 
						seal_session_g->group);
			}

		} else {		// invalid group name (strlen == 0)
			UI_print_error(ui_engine, UI_ERR_GROUP_NOT_VALID, NULL);
		}
	}else {
		UI_print_error(ui_engine, UI_ERR_GROUP_JOINED, seal_session_g->group);
	}
}		

/*
 * Callback function: member_frame -- handle the leave event
 */
void UI_gui_leave(GtkWidget* widget, UI_control_p_t ui_engine) {
	char* print_line[] = {"Status: You have successfully left a group."};

	if (ui_engine->state == UI_GROUP_IN) {
		// cancel the threads
		pthread_cancel(ui_engine->poll_member_thread);
		pthread_cancel(ui_engine->msg_recv_thread);

		// join the threads
		pthread_join(ui_engine->poll_member_thread, NULL);
		pthread_join(ui_engine->msg_recv_thread, NULL);

		// leave the group	
		SEAL_leave(seal_session_g);
	
		// reset the components
		gtk_clist_clear(GTK_CLIST(ui_engine->msg_clist));
		gtk_clist_clear(GTK_CLIST(ui_engine->member_clist));
		gtk_clist_clear(GTK_CLIST(ui_engine->status_clist));
		gtk_entry_set_editable(GTK_ENTRY(ui_engine->member_group_entry), 
				TRUE);	
		gtk_entry_set_text(GTK_ENTRY(ui_engine->member_group_entry), "");	

		// set the state
		ui_engine->state = UI_GROUP_OUT;
				
		// report status
		gtk_clist_prepend(GTK_CLIST(ui_engine->status_clist), 
				print_line);
	} else {
		UI_print_error(ui_engine, UI_ERR_GROUP_LEFT, seal_session_g->group);
	}
}		


/*
 * Callback function: exit
 */
void UI_gui_exit() {

	g_print("Thank you for using the program! Bye Bye!!\n");

	gtk_main_quit();
}


/*
 * Poll for membership events (the function is used in GUI mode only)
 */
void* UI_gui_poll_members(UI_control_p_t ui_engine) {
    pthread_cond_t cond;                // dummy condition variable
    pthread_mutex_t lock;               // dummy lock

    struct timeval tv;                  // time structures
    struct timespec ts;

    char** curr_guids;                  // current guids
    int no_curr;                        // no of current guids
    char tmp_guid[SEAL_GUID_LEN];       // temp guid

    int i, j;                           // loop counters

    // set cancel state
    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
    pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);

    // initialize the variables
    pthread_cond_init(&cond, NULL);
    pthread_mutex_init(&lock, NULL);
    pthread_mutex_lock(&lock);
    gettimeofday(&tv, NULL);        // get the current time
    ts.tv_sec = tv.tv_sec;
    ts.tv_nsec = tv.tv_usec * 1000;

    while (1) {

        // malloc 
        curr_guids = (char**)malloc(SEAL_MAX_GUID * sizeof(char*));
        for (i=0; i<SEAL_MAX_GUID; ++i) {
            curr_guids[i] = (char*)malloc(SEAL_GUID_LEN);
        }

        // get the current membership
        SEAL_read_membership(seal_session_g, curr_guids, &no_curr, NULL, NULL,
                NULL, NULL);

        // sort the current membership according to their guid
        for (i=0; i<no_curr-1; ++i) {
            for (j=0; j<no_curr-1-i; ++j) {
                if (memcmp(curr_guids[j], curr_guids[j+1], SEAL_GUID_LEN) > 0) {
                    memcpy(tmp_guid, curr_guids[j], SEAL_GUID_LEN);
                    memcpy(curr_guids[j], curr_guids[j+1], SEAL_GUID_LEN);
                    memcpy(curr_guids[j+1], tmp_guid, SEAL_GUID_LEN);
                }
            }
        }

        // print out the membership
        UI_print_members(ui_engine, curr_guids, no_curr);

        // free the resources
        for (i=0; i<SEAL_MAX_GUID; ++i) {
            free(curr_guids[i]);
        }
        free(curr_guids);

        // set the next running schedule (after 10 seconds)
        ts.tv_sec = ts.tv_sec + 10;
        pthread_cond_timedwait(&cond, &lock, &ts);
    }
    return NULL;
}

#else		// end: #ifdef GUI_ENABLED


/*-------------------------------------------------------------------------
 *						TEXTMODE FUNCTIONS
 *-------------------------------------------------------------------------*/

#ifdef SIGNATURE

/*---------------------------
 * Signature mode for Textmode
 *---------------------------*/
/*
 * Check the password
 */
void UI_textmode_passwd_check() {
	char password[500];			
	struct termios init_attr, new_attr;	// initial and new terms' attributes
	int is_done = 0;

	// turn off ECHO in the terminal
	if (tcgetattr(fileno(stdin), &init_attr) < 0) {
		fprintf(stderr, "ERROR: cannot get terminal attributes\n");
		return;
	}
	memcpy(&new_attr, &init_attr, sizeof(struct termios));
	new_attr.c_lflag &= ~ECHO;
	if (tcsetattr(fileno(stdin), TCSAFLUSH, &new_attr) < 0) {
		fprintf(stderr, "ERROR: cannot set terminal attributes\n");
		return;
	}

	// request password
	printf("Enter your password: ");
	while (!is_done) {
		memset(password, 0, sizeof(password));	
		fgets(password, sizeof(password), stdin);
		password[strlen(password)-1] = '\0';

		// check against password
		if (SEAL_set_passwd(seal_session_g, password)) {	// success
			// set the is_done flag		
			is_done = 1;	
		} else {
			// re ask the password	
			printf("\nInvalid password. Enter again:");
		}
	}

	// reset the terminal attributes
	if (tcsetattr(fileno(stdin), TCSANOW, &init_attr) < 0) {
		fprintf(stderr, "ERROR: cannot reset the terminal\n");
		return;
	}
}

#endif	// end: ifdef signature flag

/*
 * split the input into arguments
 */
int UI_textmode_split_arg(char const* input, char** arg) {
	int i;                  // loop index
	int no_arg;             // no of arguments
	int c;                  // character counter in arg

	if (input[0] == '\n') {
		return 0;
	}
	no_arg = c = 0;
	for (i = 0; input[i] != '\n' && i < UI_MAX_CHAR; ++i) {
		if (input[i] == ' ') {
			arg[no_arg][c] = '\0';
			c = 0;
			no_arg++;
			if (no_arg >= UI_MAX_ARG) {
				return no_arg;
			}
		} else {
			arg[no_arg][c++] = input[i];
		}
	}
	return ++no_arg;
}

/*
 * handle the join command
 */
void UI_textmode_join(UI_control_p_t ui_engine, int no_arg, char* const* arg) {
	char group[SEAL_GROUP_LEN];

	if (ui_engine->state == UI_GROUP_OUT) {
		if (no_arg != 2) {
			printf("Syntax: join <group_name>\n");
		} else {
			// set the group name
			memset(group, 0, SEAL_GROUP_LEN);
			strncpy(group, arg[1], SEAL_GROUP_LEN);

			// join the group
			if (SEAL_join(seal_session_g, group)) {	
				// succeed in joining the group
				ui_engine->state = UI_GROUP_IN;
				
				// start the msg recv thread
				pthread_create(&(ui_engine->msg_recv_thread), NULL, 
						(void*)&UI_receive_handler, ui_engine);

				// report status
				printf("Status: You have successfully joined a group.\n");
			} else {
				UI_print_error(ui_engine, UI_ERR_GROUP_NOT_VALID, group);
			}
		}
	} else {
		UI_print_error(ui_engine, UI_ERR_GROUP_JOINED, seal_session_g->group);
	}
}

/*
 * handle the leave command (arg is not processed here)
 */
void UI_textmode_leave(UI_control_p_t ui_engine, int no_arg, char *const* arg) {
	
	if (ui_engine->state == UI_GROUP_IN) {
		// cancel the thread
		pthread_cancel(ui_engine->msg_recv_thread);

		// join the thread
		pthread_join(ui_engine->msg_recv_thread, NULL);

		// leave the group	
		SEAL_leave(seal_session_g);
	
		// set the state	
		ui_engine->state = UI_GROUP_OUT;
		
		// report status
		printf("Status: You have successfully left a group.\n");
	} else {
		UI_print_error(ui_engine, UI_ERR_GROUP_LEFT, seal_session_g->group);
	}
}

/*
 * Handle the member command (arg is not processed here)
 */
void UI_textmode_member(UI_control_p_t ui_engine, int no_arg, char*const* arg) {

	char** curr_guids;                  // current guids
	int no_curr;                        // no of current guids
	char tmp_guid[SEAL_GUID_LEN];       // temp guid

	int i, j;                           // loop counters

	if (ui_engine->state == UI_GROUP_IN) {
			
		// malloc 
		curr_guids = (char**)malloc(SEAL_MAX_GUID * sizeof(char*));
		for (i=0; i<SEAL_MAX_GUID; ++i) {
			curr_guids[i] =
				(char*)malloc(SEAL_GUID_LEN);
		}
		
		// get the current membership
		SEAL_read_membership(seal_session_g, curr_guids, &no_curr,
				NULL, NULL, NULL, NULL);
		
		// sort the current membership according to their guid
		for (i=0; i<no_curr-1; ++i) {
			for (j=0; j<no_curr-1-i; ++j) {
				if (memcmp(curr_guids[j], curr_guids[j+1], SEAL_GUID_LEN) > 0) {
					memcpy(tmp_guid, curr_guids[j], SEAL_GUID_LEN);
					memcpy(curr_guids[j], curr_guids[j+1], SEAL_GUID_LEN);
					memcpy(curr_guids[j+1], tmp_guid, SEAL_GUID_LEN);
				}
			}
		}
		
        // print out the membership
		UI_print_members(ui_engine, curr_guids, no_curr); 

		// free the resources
		for (i=0; i<SEAL_MAX_GUID; ++i) {
			free(curr_guids[i]);
		}
		free(curr_guids);
		
	} else {
		UI_print_error(ui_engine, UI_ERR_GROUP_LEFT, seal_session_g->group);
	}
}

/*
 * handle the msg command (arg is not processed here)
 */
void UI_textmode_msg(UI_control_p_t ui_engine, char const* input) {
	char plain_msg[UI_MAX_MSG_SIZE];

	if (ui_engine->state == UI_GROUP_IN) {
		// prepare the message
		memset(plain_msg, 0, UI_MAX_MSG_SIZE);
		strncpy(plain_msg, input, UI_MAX_MSG_SIZE);
	
		// chop the newline
		if (plain_msg[strlen(plain_msg)-1] == '\n') {
			plain_msg[strlen(plain_msg)-1] = 0;
		}

		// send message
		if (SEAL_send(seal_session_g, plain_msg, strlen(plain_msg)) < 0) {
			UI_print_error(ui_engine, UI_ERR_MSG_NOT_ENCRYPTED, NULL);
		}

	} else {
		UI_print_error(ui_engine, UI_ERR_GROUP_LEFT, seal_session_g->group);
	}
}
					
/*
 * handle the exit command (arg is not processed here)
 */
void UI_textmode_exit(int no_arg, char* const* arg) {
	printf("Thank you for using the program! Bye Bye!!\n");
	exit(0);
}

#endif					// end: TEXTMODE
